"""Entity linking
This step links the entity mentions in the question to the entities in the Wikidata knowledge graph.
It inference on the REL endpoint.
"""
import argparse
import requests
import socket
from pathlib import Path

import srsly
from tqdm import tqdm

from wikimapper import WikiMapper


def socket_reachable(host, port):
    """
    Check if a socket is reachable
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(2) # set a timeout value for the socket
        s.connect((host, port))
        s.close()
        return True
    except Exception as err:
        print(err)
        return False


def main(args):
    try:
        host = args.el_endpoint.split(':')[-2].split('/')[-1]
        port = int(args.el_endpoint.split(':')[-1])
    except Exception as exc:
        raise ValueError(f"Can't parse the endpoint {args.el_endpoint}") from exc
    if not socket_reachable(host, port):
        raise RuntimeError(f"Can't reach the endpoint {args.el_endpoint}")
    mapper = WikiMapper(args.wikimapper_db)

    with open(args.input, "r", encoding="utf-8") as f:
        total_lines = len(f.readlines())
    questions = srsly.read_jsonl(args.input)
    cnt_qid_not_found = 0
    cnt_total_ground = 0
    all_linked = []
    for question in tqdm(questions, total=total_lines, desc=f"Entity linking {args.input}"):
        document = {
            'text': question[args.ground_on],
        }
        api_results = requests.post(args.el_endpoint, json=document, timeout=60).json()
        cnt_total_ground += len(api_results)
        qids = []
        spans = []
        entities = []
        for result in api_results:
            start_pos, mention_length, mention, entity, disambiguation_cofidence, mention_detection_confidence, tag = result
            qid = mapper.title_to_id(entity)
            span = (start_pos, start_pos + mention_length)
            if qid is None:
                cnt_qid_not_found += 1
            else:
                qids.append(qid)
                entities.append(entity)
                spans.append(span)
        linked = {
            "question": question[args.ground_on],
            "question_entities": qids,
            "spans": spans,
            "entity_names": entities
        }
        if "id" in question:
            linked["id"] = question["id"]
        all_linked.append(linked)
    print(f"{cnt_qid_not_found} / {cnt_total_ground} grounded entities not converted to Wikidata qids")
    # check whether the folder exists
    folder_path = Path(args.output).parent
    if not folder_path.exists():
        folder_path.mkdir(parents=True)
        print(f"Folder {folder_path} created")
    Path(args.output).parent.mkdir(parents=True, exist_ok=True)
    srsly.write_jsonl(args.output, all_linked)
    print(f"Entity linking result saved to {args.output}")


def add_arguments(parser):
    """Add entity linking arguments to the parser"""
    parser.description = '''Entity linking on Wikidata.
    The input is a jsonl file. The field of interest is specified by the argument --ground-on.
    The output is a jsonl file, each line is a dict with keys: id, question_entities, spans, entity_names.
    '''
    parser.add_argument('-i', '--input', type=str, help='Input file path, in which the question is stored')
    parser.add_argument('-o', '--output', type=str, help='Output file path, in which the entity linking result is stored')
    parser.add_argument('-e', '--el-endpoint', type=str, default='http://127.0.0.1:1235', help='REL endpoint')
    parser.add_argument('--wikimapper-db', type=str, default='resources/wikimapper/index_enwiki.db', help='Wikimapper database path')
    parser.add_argument('--ground-on', type=str, default='question', help='The key to ground on, the corresponding text will be sent to the REL endpoint for entity linking')
    

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()
    main(args)



from functools import lru_cache
from SPARQLWrapper import SPARQLWrapper, JSON

from .graph_base import KnowledgeGraphBase


class Wikidata(KnowledgeGraphBase):
    PREFIXES: str = """PREFIX wd: <http://www.wikidata.org/entity/>
        PREFIX wds: <http://www.wikidata.org/entity/statement/>
        PREFIX wdv: <http://www.wikidata.org/value/>
        PREFIX wdt: <http://www.wikidata.org/prop/direct/>
        PREFIX wikibase: <http://wikiba.se/ontology#>
        PREFIX p: <http://www.wikidata.org/prop/>
        PREFIX ps: <http://www.wikidata.org/prop/statement/>
        PREFIX pq: <http://www.wikidata.org/prop/qualifier/>
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        PREFIX bd: <http://www.bigdata.com/rdf#>
        """
    ENTITY_PREFIX: str = "http://www.wikidata.org/entity/Q"

    def __init__(self, endpoint, prepend_prefixes=False, exclude_qualifiers=True):
        """Create a Wikidata query handler.

        Args:
            endpoint (str): SPARQL endpoint, e.g. https://query.wikidata.org/sparql
                Note that the protocal part (like https) is necessary.
            prepend_prefixes (bool, optional): whether to prepend prefixes to the query.
                Necessary for endpoints without pre-defined prefixes. Defaults to False.
            exclude_qualifiers (bool, optional): whether to filter out qualifiers in the
                queried entities. If set to True, only Wikidata entities (QXX) will be
                considered. Defaults to True.
        """
        self.sparql = SPARQLWrapper(endpoint)
        self.sparql.setReturnFormat(JSON)
        self.prepend_prefixes = prepend_prefixes
        self.exclude_qualifiers = exclude_qualifiers
        self.name = 'wikidata'

    def queryWikidata(self, query):
        if self.prepend_prefixes:
            query = self.PREFIXES + query

        self.sparql.setQuery(query)
        try:
            ret = self.sparql.queryAndConvert()
            result = ret['results']['bindings']
        except Exception as exeption:
            print(f'Failed executing query: {query}')
            print(f'Exception: {exeption}')
            result = []
        return result

    @staticmethod
    def get_pid_from_uri(uri):
        """Get property id from uri."""
        return uri.split('/')[-1]

    @staticmethod
    def is_qid(qid):
        """Check if qid is a valid Wikidata entity id."""
        return qid.startswith('Q') and qid[1:].isdigit()

    @staticmethod
    def is_pid(pid):
        """Check if pid is a valid Wikidata property id."""
        return pid.startswith('P') and pid[1:].isdigit()

    def get_quantifier_filter(self, var_name):
        """Get quantifier filter string where the var is restricted to be entities.
        If exclude_qualifiers is set to False, return empty string.

        Note: in Wikidata, entities are prefixed with "http://www.wikidata.org/entity/Q",
            while qualifiers are non-entity (and mostly string) values.
        """
        return f'FILTER(STRSTARTS(STR(?{var_name}), "{self.ENTITY_PREFIX}"))' if self.exclude_qualifiers else ''

    def search_one_hop_relations(self, src, dst):
        """Search one hop relation between src and dst.
        
        Args:
            src (str): source entity
            dst (str): destination entity
        
        Returns:
            list[list[str]]: list of paths, each path is a list of PIDs
        """
        if not self.is_qid(src) or not self.is_qid(dst):
            return []

        query = f"""
            SELECT DISTINCT ?r WHERE {{
                wd:{src} ?r wd:{dst}.
            }}
            """
        paths = self.queryWikidata(query)
        # Keep only PIDs in the paths
        paths = [[self.get_pid_from_uri(path['r']['value'])] for path in paths]
        return paths

    def search_two_hop_relations(self, src, dst):
        """Search two hop relation between src and dst.
        
        Args:
            src (str): source entity
            dst (str): destination entity
        
        Returns:
            list[list[str]]: list of paths, each path is a list of PIDs
        """
        if not self.is_qid(src) or not self.is_qid(dst):
            return []

        query = f"""
            SELECT DISTINCT ?r1 ?r2 WHERE {{
                wd:{src} ?r1 ?x.
                ?x ?r2 wd:{dst}.
                {self.get_quantifier_filter('x')}
            }}
            """
        paths = self.queryWikidata(query)
        # Keep only PIDs in the paths
        paths = [[self.get_pid_from_uri(path['r1']['value']),
                  self.get_pid_from_uri(path['r2']['value'])]
                 for path in paths]
        return paths

    @lru_cache
    def deduce_leaves(self, src, path, limit=2000):
        """Deduce leave entities from source entity following the path.
        
        Args:
            src_entity (str): source entity
            path (tuple[str]): path from source entity to destination entity
            limit (int, optional): limit of the number of leaves. Defaults to 2000.
        
        Returns:
            list[str]: list of leaves. Each leaf is a QID.
        """
        if len(path) >= 3:
            raise NotImplementedError(f'Currenly only support paths with length less than 3, got {len(path)}')
        if not self.is_qid(src):
            return []

        if len(path) == 0:
            return [src]
        if len(path) == 1:
            query = f"""
                SELECT DISTINCT ?x WHERE {{
                    wd:{src} wdt:{path[0]} ?x.
                    {self.get_quantifier_filter('x')}
                    }}
                LIMIT {limit}
            """
        else: # len(path) == 2
            query = f"""
                SELECT DISTINCT ?x WHERE {{
                    wd:{src} wdt:{path[0]} ?y.
                    ?y wdt:{path[1]} ?x.
                    {self.get_quantifier_filter('y')}
                    {self.get_quantifier_filter('x')}
                }}
                LIMIT {limit}
            """
        if self.prepend_prefixes:
            query = self.PREFIXES + query
        leaves = self.queryWikidata(query)
        # Keep only QIDs in the leaves
        leaves = [leaf['x']['value'].split('/')[-1] for leaf in leaves]
        return leaves

    def deduce_leaves_from_multiple_srcs(self, srcs, path, limit=2000):
        """Deuce leave entities from multiple source entities following the path.

        Args:
            srcs (list[str]): list of source entities
            path (list[str]): path from source entity to destination entity
            limit (int, optional): limit of the number of leaves. Defaults to 200.

        Returns:
            list[str]: list of leaves. Each leaf is a QID.
        """
        if len(path) >= 2:
            raise NotImplementedError(f'Currenly only support paths with length less than 2, got {len(path)}')
        if len(path) == 0:
            return srcs
        srcs = [src for src in srcs if self.is_qid(src)]
        if len(srcs) == 0:
            return []

        query = f"""
            SELECT DISTINCT ?x WHERE {{
                VALUES ?src {{wd:{' wd:'.join(srcs)}}}
                ?src wdt:{path[0]} ?x.
                {self.get_quantifier_filter('x')}
            }}
            LIMIT {limit}
            """
        if self.prepend_prefixes:
            query = self.PREFIXES + query
        leaves = self.queryWikidata(query)
        # Keep only QIDs in the leaves
        leaves = [leaf['x']['value'].split('/')[-1] for leaf in leaves]
        return leaves

    @lru_cache
    def get_neighbor_relations(self, src, hop=1, limit=100):
        """Get all relations connected to an entity. The relations are
        limited to direct relations (those with wdt: prefix).

        Args:
            src (str): source entity
            hop (int, optional): hop of the relations. Defaults to 1.
            limit (int, optional): limit of the number of relations. Defaults to 100.

        Returns:
            list[str] | list[tuple(str,)]: list of relations. Each relation is a PID or a tuple of PIDs.
        """
        if hop >= 3:
            raise NotImplementedError(f'Currenly only support relations with hop less than 3, got {hop}')
        if not self.is_qid(src):
            return []

        if hop == 1:
            query = f"""SELECT DISTINCT ?rel WHERE {{
                wd:{src} ?rel ?obj .
                FILTER(REGEX(STR(?rel), "^http://www.wikidata.org/prop/direct/"))
                {self.get_quantifier_filter('obj')}
                }}
                LIMIT {limit}
                """
        else: # hop == 2
            query = f"""SELECT DISTINCT ?rel1 ?rel2 WHERE {{
                wd:{src} ?rel1 ?obj1 .
                ?obj1 ?rel2 ?obj2 .
                FILTER(REGEX(STR(?rel1), "^http://www.wikidata.org/prop/direct/"))
                FILTER(REGEX(STR(?rel2), "^http://www.wikidata.org/prop/direct/"))
                {self.get_quantifier_filter('obj1')}
                {self.get_quantifier_filter('obj2')}
                }}
                LIMIT {limit}
                """
        if self.prepend_prefixes:
            query = self.PREFIXES + query

        relations = self.queryWikidata(query)
        if hop == 1:
            relations = [self.get_pid_from_uri(relation['rel']['value'])
                         for relation in relations]
        else:
            relations = [(self.get_pid_from_uri(relation['rel1']['value']),
                          self.get_pid_from_uri(relation['rel2']['value']))
                         for relation in relations]
        return relations

    @lru_cache
    def get_label(self, identifier):
        """Get label of an entity or a relation. If no label is found, return None.

        Args:
            identifier (str): entity or relation, a QID or a PID

        Returns:
            str | None: label of the entity or relation
        """
        if not self.is_qid(identifier) and not self.is_pid(identifier):
            return identifier

        query = f"""
            SELECT ?label
                WHERE {{
                    BIND(wd:{identifier} AS ?identifier)
                    ?identifier rdfs:label ?label .
                    FILTER(LANG(?label) = "en")
                    }}
                LIMIT 1
            """
        if self.prepend_prefixes:
            query = self.PREFIXES + query
        label = self.queryWikidata(query)
        if len(label) == 0:
            print(f'No label for identifier {identifier}.')
            return None
        label = label[0]['label']['value']
        return label



from functools import lru_cache
from SPARQLWrapper import SPARQLWrapper, JSON

from .graph_base import KnowledgeGraphBase


class Freebase(KnowledgeGraphBase):
    PREFIXES: str = """
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        PREFIX ns: <http://rdf.freebase.com/ns/>
        """

    def __init__(self, endpoint, prepend_prefixes=True) -> None:
        self.sparql = SPARQLWrapper(endpoint)
        self.sparql.setReturnFormat(JSON)
        self.prepend_prefixes = prepend_prefixes
        self.name = 'freebase'

    def queryFreebase(self, query):
        if self.prepend_prefixes:
            query = self.PREFIXES + query

        self.sparql.setQuery(query)
        try:
            ret = self.sparql.queryAndConvert()
            result = ret['results']['bindings']
        except Exception as exeption:
            print(exeption)
            print(f'Failed executing query: {query}')
            result = []
        return result

    @staticmethod
    def get_id_from_uri(uri):
        """Get id from uri."""
        return uri.split('/')[-1]

    @lru_cache
    def search_one_hop_relations(self, src, dst):
        """Search one hop relation between src and dst.
        
        Args:
            src (str): source entity
            dst (str): destination entity
        
        Returns:
            list[list[str]]: list of paths, each path is a list of PIDs
        """
        query = f"""
            SELECT distinct ?r1 where {{
                ns:{src} ?r1_ ns:{dst} . 
                FILTER REGEX(?r1_, "http://rdf.freebase.com/ns/")
                BIND(STRAFTER(STR(?r1_),str(ns:)) AS ?r1)
            }}
        """
        paths = self.queryFreebase(query)
        paths = [[path['r1']['value']] for path in paths]
        return paths

    @lru_cache
    def search_two_hop_relations(self, src, dst):
        query = f"""
            SELECT distinct ?r1 ?r2 where {{
                ns:{src} ?r1_ ?e1 . 
                ?e1 ?r2_ ns:{dst} .
                FILTER REGEX(?e1, "http://rdf.freebase.com/ns/")
                FILTER REGEX(?r1_, "http://rdf.freebase.com/ns/")
                FILTER REGEX(?r2_, "http://rdf.freebase.com/ns/")
                FILTER (?r1_ != <http://rdf.freebase.com/ns/type.object.type>)
                FILTER (?r2_ != <http://rdf.freebase.com/ns/type.object.type>)
                BIND(STRAFTER(STR(?r1_),str(ns:)) AS ?r1)
                BIND(STRAFTER(STR(?r2_),str(ns:)) AS ?r2)
            }}
        """
        paths = self.queryFreebase(query)
        paths = [[path['r1']['value'], path['r2']['value']] for path in paths]
        return paths

    @lru_cache
    def deduce_leaves(self, src, path, limit=2000):
        """Deduce leave entities from source entity following the path.
        
        Args:
            src_entity (str): source entity
            path (tuple[str]): path from source entity to destination entity
            limit (int, optional): limit of the number of leaves. Defaults to 2000.
        
        Returns:
            list[str]: list of leaves. Each leaf is a QID.
        """
        if len(path) >= 3:
            raise NotImplementedError(f'Currenly only support paths with length less than 3, got {len(path)}')
        if len(path) == 0:
            return [src]
        if len(path) == 1:
            query = f"""
                SELECT DISTINCT ?leaf WHERE {{
                    ns:{src} ns:{path[0]} ?t0_ .
                    FILTER REGEX(?t0_, "http://rdf.freebase.com/ns/")
                    BIND(STRAFTER(STR(?t0_),STR(ns:)) AS ?leaf)
                }} LIMIT {limit}
                """
        else: # len(path) == 2:
            query = f"""
                SELECT DISTINCT ?leaf WHERE {{
                    ns:{src} ns:{path[0]} ?e1_ . 
                    ?e1_ ns:{path[1]} ?e2_ .
                    FILTER REGEX(?e1_, "http://rdf.freebase.com/ns/")
                    FILTER REGEX(?e2_, "http://rdf.freebase.com/ns/")
                    BIND(STRAFTER(STR(?e2_),str(ns:)) AS ?leaf)
                }} LIMIT {limit}
                """
        results = self.queryFreebase(query)
        return [i['leaf']['value'] for i in results]

    def deduce_leaves_from_multiple_srcs(self, srcs, path, limit=2000):
        """Deuce leave entities from multiple source entities following the path.

        Args:
            srcs (list[str]): list of source entities
            path (list[str]): path from source entity to destination entity
            limit (int, optional): limit of the number of leaves. Defaults to 200.

        Returns:
            list[str]: list of leaves. Each leaf is a QID.
        """
        if len(path) >= 2:
            raise NotImplementedError(f'Currenly only support paths with length less than 2, got {len(path)}')
        if len(path) == 0:
            return srcs
        query = f"""
            SELECT DISTINCT ?leaf WHERE {{
                VALUES ?src {{ns:{' ns:'.join(srcs)}}}
                ?src ns:{path[0]} ?t0_ .
                FILTER REGEX(?t0_, "http://rdf.freebase.com/ns/")
                BIND(STRAFTER(STR(?t0_),STR(ns:)) AS ?leaf)
            }} LIMIT {limit}
            """
        results = self.queryFreebase(query)
        return [i['leaf']['value'] for i in results]

    @lru_cache
    def get_neighbor_relations(self, src, hop=1, limit=100):
        """Get all relations connected to an entity. The relations are
        limited to direct relations (those with wdt: prefix).

        Args:
            src (str): source entity
            hop (int, optional): hop of the relations. Defaults to 1.
            limit (int, optional): limit of the number of relations. Defaults to 100.

        Returns:
            list[str] | list[tuple(str,)]: list of relations. Each relation is a PID or a tuple of PIDs.
        """
        if hop >= 3:
            raise NotImplementedError(f'Currenly only support relations with hop less than 3, got {hop}')
        if hop == 1:
            query = f"""
                SELECT DISTINCT ?rel WHERE {{
                    ns:{src} ?r0_ ?t0 .
                    FILTER REGEX(?r0_, "http://rdf.freebase.com/ns/")
                    FILTER REGEX(?t0, "http://rdf.freebase.com/ns/")
                    FILTER (?r0_ != <http://rdf.freebase.com/ns/type.object.type>)
                    BIND(STRAFTER(STR(?r0_),STR(ns:)) AS ?rel)
                }} LIMIT {limit}
                """
        elif hop == 2:
            query = f"""
                SELECT DISTINCT ?rel0, ?rel1 WHERE {{     
                    {{
                        SELECT DISTINCT ?t0, ?rel0 WHERE {{
                        ns:{src} ?r0_ ?t0 .
                        FILTER REGEX(?r0_, "http://rdf.freebase.com/ns/")
                        FILTER REGEX(?t0, "http://rdf.freebase.com/ns/")
                        FILTER (?r0_ != <http://rdf.freebase.com/ns/type.object.type>)
                        BIND(STRAFTER(STR(?r0_),STR(ns:)) AS ?rel0)
                        }} LIMIT 10
                    }}
                    ?t0 ?r1_ ?t1 .
                    FILTER REGEX(?r1_, "http://rdf.freebase.com/ns/")
                    FILTER REGEX(?t1, "http://rdf.freebase.com/ns/")
                    FILTER (?r1_ != <http://rdf.freebase.com/ns/type.object.type>)
                    BIND(STRAFTER(STR(?r1_),STR(ns:)) AS ?rel1)
                }} LIMIT {limit}
                """
        results = self.queryFreebase(query)
        if hop == 1:
            paths = [path['rel']['value'] for path in results]
        else:
            paths = [(path['rel0']['value'], path['rel1']['value'])
                     for path in results]
        return paths

    @lru_cache
    def get_label(self, identifier):
        query = f"""
            SELECT ?label
            WHERE {{
                ns:{identifier} rdfs:label ?label .
                FILTER (langMatches(lang(?label), "EN"))
            }} LIMIT 1
            """
        results = self.queryFreebase(query)
        return results[0]['label']['value'] if results else None

    def get_relation_label(self, relation):
        """For freebase, relation label is the same as the relation identifier."""
        return relation



"""5. Negative Sampling

Regarding negative sampling method, the author states in the paper:
> We replace the observed relation at each time step with other sampled relations as the
negative instances to optimize the probability of the observed ones.

e.g.
python preprocess/negative_sampling.py \
    --scored-path-file data/preprocess/paths_scored.jsonl \
    --output-file data/preprocess/train_.jsonl\
    --positive-threshold 0.3
"""
import os
import sys
import argparse
import random
from collections import defaultdict
from functools import lru_cache

import srsly
from tqdm import tqdm

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', )))
from knowledge_graph import Wikidata, Freebase, KnowledgeGraphBase

END_REL = "END OF HOP"


def sample_negative_relations(soruce_entities, prev_path, positive_connections,
                              num_negative, kg: KnowledgeGraphBase):
    """A helper function to sample negative relations.
    
    Args:
        soruce_entities (list[str]): list of source entities
        prev_path (list[str]): previous path / relations
        positive_connections (dict): a dictionary of positive connections
        num_negative (int): number of negative relations to sample
        kg (KnowledgeGraphBase): a knoledge graph instance
        
    Returns:
        list[str]: list of negative relations
    """
    negative_relations = set()
    for src in soruce_entities:
        # get all relations connected to current tracked entities (question or intermediate entities)
        negative_relations |= set(kg.get_neighbor_relations(src, limit=50))
        if len(negative_relations) > 100:  # yet another magic number :(
            break
    negative_relations = negative_relations - positive_connections[tuple(prev_path)]
    if len(negative_relations) == 0:
        return []
    negative_relations = random.choices(list(negative_relations), k=num_negative)
    return negative_relations


def is_candidate_space_too_large(path, question_entities, kg, candidate_depth_multiplier=5):
    """Check whether the number of the candidate entities along the path is too large.
    
    Args:
        path (list[str]): path from source entity to destination entity
        question_entities (list[str]): list of question entities
        kg (KnowledgeGraphBase): a knowledge graph instance
        candidate_depth_multiplier (int, optional): a multiplier to control the number of candidate entities
            at each depth. Defaults to 10.
    """
    flag_too_large = False
    for i in range(1, len(path)):
        prev_path = tuple(path[:i])
        # The further the path is from the question, the greater the search space becomes.
        limit = candidate_depth_multiplier ** i
        candidate_entities = set()
        for src in question_entities:
            candidate_entities |= set(kg.deduce_leaves(src, prev_path, limit=limit))
            # Stop early if the number of candidate entities is already very large
            if len(candidate_entities) > limit:
                break
        # Check the entity space at each depth, if it is too large at any depth, we flag the path as too large.
        if len(candidate_entities) > limit:
            flag_too_large = True
            break
    return flag_too_large


def sample_records_from_path(path, question, question_entities, positive_connections,
                             kg, num_negative):
    """Sample training records from a path.
    
    Returns:
        list[dict]: list of training records, each record has the following fields:
            - question (str): the question
            - prev_path (list): previous relations up to a relation (positive_relation) in the path
            - positive_relation (str): the next relation of the prev_path is regarded as the positive relation
            - negative_relations (list): a list of negative relations, the number is specified by num_negative
    """
    # My interpretation: If the number of candidate entities is too large, we simply discard this path.
    # But isn't it weird? The author checks the number of connected entities to the question entities
    # with each relation along the path, and simply discard those paths with too many connected entities.
    if is_candidate_space_too_large(path, question_entities, kg, candidate_depth_multiplier=5):
        return []

    path = path + [END_REL]

    records = []
    tracked_entities = question_entities
    for i, current_relation in enumerate(path):
        prev_path = path[:i]
        negative_relations = sample_negative_relations(tracked_entities, prev_path, positive_connections,
                                                       num_negative, kg)
        if len(negative_relations) == 0:
            continue
        record = {
            'question': question,
            'prev_path': prev_path,
            'positive_relation': current_relation,
            'negative_relations': negative_relations
        }
        records.append(record)
        if current_relation != END_REL:
            # update tracked entities
            tracked_entities = kg.deduce_leaves_from_multiple_srcs(tracked_entities, [current_relation], limit=100)
    return records


def get_positive_connections_along_paths(paths):
    """Collect positive connections along paths. A positive connection is defined as
    {prev_relations: next_relation}. END_REL is added to the end of each path.
    
    Returns:
        dict: a dictionary of positive connections
    """
    positive_connections = defaultdict(set)
    for path in paths:
        path = path + [END_REL]
        for i, rel in enumerate(path):
            positive_connections[tuple(path[:i])].add(rel)
    return positive_connections


def convert_records_relation_id_to_lable(records, kg):
    """Convert relation ids to relation labels in each record.
    """
    processed_records = []

    @lru_cache
    def get_label(rel):
        if rel == END_REL:
            return END_REL
        return kg.get_relation_label(rel) or rel

    for record in tqdm(records, desc='Converting relation ids to labels'):
        record['prev_path'] = [get_label(rel) for rel in record['prev_path']]
        record['positive_relation'] = get_label(record['positive_relation'])
        record['negative_relations'] = [get_label(rel) for rel in record['negative_relations']]
        processed_records.append(record)
    return processed_records


def create_jsonl_dataset(records):
    """It combines the question and prev_path to query. Each train sample is a dict with the following fields:
    - query (str): question + prev_path
    - positive (str): the next relation of the prev_path is regarded as the positive relation
    - negatives (list): a list of negative relations
    
    Args:
        records (list[dict]): list of records
        
    Returns:
        list[dict]: list of train samples
    """
    samples = []
    for record in records:
        sample = {
            'query': record['question'] + ' [SEP] ' + ' # '.join(record['prev_path']),
            'positive': record['positive_relation'],
            'negatives': record['negative_relations']
        }
        samples.append(sample)
    return samples


def main(args):
    if args.knowledge_graph == 'freebase':
        kg = Freebase(args.sparql_endpoint)
    else:
        kg = Wikidata(args.sparql_endpoint)
    positive_threshold = args.positive_threshold
    # Each sample has the following fields:
    # - id: sample id
    # - question: question text
    # - question_entities: list of question entities (ids)
    # - answer_entities: list of answer entities (ids)
    # - question: question text
    # - paths: list of paths
    # - path_scores: list of path scores
    samples = srsly.read_jsonl(args.scored_path_file)
    total = sum(1 for _ in srsly.read_jsonl(args.scored_path_file))
    train_records = []
    for sample in tqdm(samples, total=total, desc='Negative sampling'):
        paths = sample['paths']
        path_scores = sample['path_scores']
        question = sample['question']
        question_entities = sample['question_entities']

        # Filter out paths with low scores.
        paths = [path for path, score in zip(paths, path_scores) if score >= positive_threshold]
        path_scores = [score for score in path_scores if score >= positive_threshold]
        if len(paths) != len(path_scores):
            raise ValueError(f'The number of paths and path scores are not equal. {len(paths)} != {len(path_scores)}')

        # A dictionary of {prev_rels: {next_rel, next_rel, ...}, ...},
        # where prev_rels is a tuple of previous relations.
        positive_connections = get_positive_connections_along_paths(paths)

        for path in paths:
            train_records.extend(sample_records_from_path(path, question, question_entities,
                                                          positive_connections, kg, args.num_negative))
    print(f"Number of training records: {len(train_records)}")
    train_records = convert_records_relation_id_to_lable(train_records, kg)
    train_records = create_jsonl_dataset(train_records)
    srsly.write_jsonl(args.output_path, train_records)
    print(f"Training samples are saved to {args.output_path}")


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--sparql-endpoint', default='http://localhost:1234/api/endpoint/sparql', help='wikidata endpoint')
    parser.add_argument('-kg', '--knowledge-graph', default='wikidata', help='knowledge graph name')
    parser.add_argument('--scored-path-file', help='The file containing scored paths')
    parser.add_argument('--output-path', help='The path to the output file')
    parser.add_argument('--positive-threshold', type=float, default=0.5, help='The threshold to determine whether a path is positive or negative')
    parser.add_argument('--num-negative', type=int, default=15, help='The number of negative relations to sample for each positive relation')
    args = parser.parse_args()

    main(args)



"""Visualize the graph (represented as a set of triplets) using pyvis.
The visualized subgraphs are html files.
"""
import os
import argparse
import json
from pathlib import Path

import srsly
from pyvis.network import Network
from tqdm import tqdm
from bs4 import BeautifulSoup as Soup

from .knowledge_graph import KnowledgeGraphBase, Wikidata, Freebase


def visualize_subgraph(sample, kg: KnowledgeGraphBase):
    """Visualize the subgraph. It returns an html string.
    """
    net = Network(directed=True, font_color='#000000')
    net.barnes_hut()
    question_entities = sample['question_entities'] if 'question_entities' in sample else []
    answer_entities = sample['answer_entities'] if 'answer_entities' in sample else []
    # Add question entities even if they are not in the triplets
    for entity in question_entities:
        net.add_node(entity, label=kg.get_label(entity), color='#114B7A')
    for triplet in sample['triplets']:
        subject, relation, obj = triplet
        subject_label = kg.get_label(subject)
        subject_options = {'color':'#114B7A'} if subject in question_entities else {}
        obj_label = kg.get_label(obj)
        obj_options = {'color':'#1B5E20'} if obj in answer_entities else {}
        relation_label = kg.get_label(relation)
        net.add_node(subject, label=subject_label, **subject_options)
        net.add_node(obj, label=obj_label, **obj_options)
        net.add_edge(subject, obj, label=relation_label)

    net_options = {
        'shape': 'dot',
        'font': {
            'size': '1em',
            'face': 'fontFace',
            'strokeColor': '#fff',
            'strokeWidth': 5
        },
        'size': '1.5em',
    }
    net.set_options(json.dumps(net_options))
    return net.generate_html(notebook=False)


def add_text_to_html(html, text):
    soup = Soup(html, 'html.parser')
    style_tag = soup.new_tag('style')
    style_tag.string = '''
        .background-text {
            position: absolute;
            z-index: 1;
            top: 0;
            left: 0;
            font-size: 2em;
            color: #ccc;
        }
    '''
    soup.head.append(style_tag)
    p_tag = soup.new_tag('p', attrs={'class': 'background-text'}, style="white-space:pre-wrap")
    p_tag.string = text
    soup.body.append(p_tag)
    return soup.prettify()
    

def main(args):
    if args.knowledge_graph == 'wikidata':
        knowledge_graph = Wikidata(args.sparql_endpoint)
    else:
        knowledge_graph = Freebase(args.sparql_endpoint)
    samples = srsly.read_jsonl(args.input)
    total = sum(1 for _ in srsly.read_jsonl(args.input))
    total = min(total, args.max_output)
    if not os.path.exists(args.output_dir):
        Path(args.output_dir).mkdir(parents=True, exist_ok=True)
        print(f'Created output directory: {args.output_dir}')
    for i, sample in enumerate(tqdm(samples, desc='Visualizing graphs', total=total)):
        if i >= args.max_output:
            break
        html = visualize_subgraph(sample, knowledge_graph)
        text_to_append = f"Question: {sample['question']}"
        if 'answer' in sample:
            text_to_append += f"\n    Answer: {sample['answer']}"
        html = add_text_to_html(html, text_to_append)
        output_path = os.path.join(args.output_dir, sample['id'] + '.html')
        with open(output_path, 'w', encoding='utf-8') as fout:
            fout.write(html)
    print(f'Visualized graphs outputted to {args.output_dir}.')


def add_arguments(parser):
    parser.description = 'Visualize the graph (represented as a set of triplets) using pyvis.'
    parser.add_argument('-i', '--input', required=True, help='The input subgraph file path.')
    parser.add_argument('-o', '--output-dir', required=True, help='The output directory path.')
    parser.add_argument('-e', '--sparql-endpoint', type=str, default='http://localhost:1234/api/endpoint/sparql',
                        help='SPARQL endpoint for Wikidata or Freebase services. In this step, it is used to get the labels of entities.\
                        (Default: http://localhost:1234/api/endpoint/sparql)')
    parser.add_argument('-kg', '--knowledge-graph', type=str, choices=('wikidata', 'freebase'), default='wikidata',
                        help='The knowledge graph type to use. (Default: wikidata)')
    parser.add_argument('--max-output', type=int, default=1000,
                        help='The maximum number of graphs to output. This is useful for debugging. (Default: 1000)')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()
    main(args)



"""Command-line interface for SRTK."""
import argparse

from .link_wikidata import add_arguments as add_link_wikidata_arguments
from .link_wikidata import main as link_wikidata
from .preprocess import add_arguments as add_preprocess_arguments
from .preprocess import main as preprocess
from .retrieve import add_arguments as add_retrieve_arguments
from .retrieve import main as retrieve
from .train import add_arguments as add_train_arguments
from .train import main as train
from .visualize import add_arguments as add_visualize_arguments
from .visualize import main as visualize


def main():
    parser = argparse.ArgumentParser(description='SRTK: A toolkit for smantic-relevant subgraph retrieval')
    subparsers = parser.add_subparsers(help='sub-command help')

    parser_link_wikidata = subparsers.add_parser('link-wikidata', help='link entities to Wikidata')
    add_link_wikidata_arguments(parser_link_wikidata)
    parser_link_wikidata.set_defaults(func=link_wikidata)

    parser_preprocess = subparsers.add_parser('preprocess', help='preprocess the data')
    add_preprocess_arguments(parser_preprocess)
    parser_preprocess.set_defaults(func=preprocess)

    parser_train = subparsers.add_parser('train', help='train a subgraph retriever')
    add_train_arguments(parser_train)
    parser_train.set_defaults(func=train)

    parser_retrieve = subparsers.add_parser('retrieve', help='retrieve a subgraph with a natural query')
    add_retrieve_arguments(parser_retrieve)
    parser_retrieve.set_defaults(func=retrieve)

    parser_visualize = subparsers.add_parser('visualize', help='visualize the retrieved subgraph')
    add_visualize_arguments(parser_visualize)
    parser_visualize.set_defaults(func=visualize)

    args = parser.parse_args()
    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_help()

if __name__ == '__main__':
    main()



import json
import os
from pathlib import Path

import lightning.pytorch as pl
import torch
import torch.nn.functional as F
from transformers import AutoModel


class LitSentenceEncoder(pl.LightningModule):
    """Lightning module """
    def __init__(self, model_name_or_path, temperature=0.07, lr=5e-5):
        super().__init__()
        self.model = AutoModel.from_pretrained(model_name_or_path)
        self.config = self.model.config
        self.loss_fn = F.cross_entropy
        self.temperature = temperature
        self.lr = lr

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    @staticmethod
    def cls_pool(last_hidden_states, attention_mask=None):
        """CLS pool the sentence embedding.
        This is the pooling method adopted by RUC's SR paper.
        
        Args:
            last_hidden_states: [..., seq_len, embedding_dim]
            attention_mask: [..., seq_len] silently ignored!
                It exists for compatibility with other pooling methods.
        Returns:
            pooled_embedding: [..., embedding_dim]
        """
        return last_hidden_states[..., 0, :]

    def compute_embedding_similarity(self, query, target):
        """Compute the similarity between query and target(s) embeddings.
        
        Args:
            query: [batch_size, 1, embedding_dim]
            target: [batch_size, k, embedding_dim]
        
        Returns:
            similarity: [batch_size, k]
        """
        return F.cosine_similarity(query, target, dim=-1) / self.temperature

    def compute_sentence_similarity(self, query, target):
        """Compute the similarity between query and target(s) sentence embeddings.
        The query & target(s) sentence embedding are first pooled. Then the similarity
        is computed between the pooled query and target(s) embeddings. 
        
        Args:
            query: [..., 1, seq_len, embedding_dim]
                query sentence embedding
            target: [..., k, seq_len, embedding_dim]
                target sentence(s) embedding

        Returns:
            similarity: [..., k]
        """        
        embeddings = torch.cat([query, target], dim=-3)  # [..., 1 + k, seq_len, embedding_dim]
        embeddings = self.cls_pool(embeddings) # [..., 1 + k, embedding_dim]
        query_embeddings = embeddings[..., 0:1, :]  # [..., 1, embedding_dim]
        samples_embeddings = embeddings[..., 1:, :]  # [..., k, embedding_dim]
        similarity = self.compute_embedding_similarity(query_embeddings, samples_embeddings)
        return similarity

    def batch_forward(self, batch):
        """The common forward function for both training and inference."""
        # In each sentence group, the first sentence is the query, the second is the positive,
        # and the rest are negatives.
        # batch = {'input_ids': input_ids, 'attention_mask': attention_mask}
        # input_ids: [batch_size, 1(query) + 1(positive) + neg (negative), seq_len]
        batch_size, n_samples, seq_len = batch['input_ids'].shape  # n_sentences = 1 + 1 + neg
        input_ids = batch['input_ids'].view(-1, seq_len)
        attention_mask = batch['attention_mask'].view(-1, seq_len)
        # outputs.last_hidden_state: [batch_size * (1 + 1 + neg), seq_len, embedding_dim]
        outputs = self.model(input_ids, attention_mask=attention_mask, return_dict=True)
        embeddings = outputs.last_hidden_state.view(batch_size, n_samples, seq_len, -1)
        query_embedding = embeddings[:, 0:1]  # [batch_size, 1, seq_len, embedding_dim]
        samples_embedding = embeddings[:, 1:]  # [batch_size, 1 + neg, seq_len, embedding_dim]
        # similarity: [batch_size, 1 + neg]
        query_samples_similarity = self.compute_sentence_similarity(query_embedding, samples_embedding)
        # The zerot-th label, where positive sample locates, is set to 0.
        labels = torch.zeros(query_samples_similarity.shape[0], dtype=torch.long,
                             device=query_samples_similarity.device)
        loss = self.loss_fn(query_samples_similarity, labels)
        return loss

    def training_step(self, batch, batch_idx):
        train_loss =  self.batch_forward(batch)
        self.log('train_loss', train_loss)
        return train_loss

    def validation_step(self, batch, batch_idx):
        val_loss =  self.batch_forward(batch)
        self.log('val_loss', val_loss)
        return val_loss

    def configure_optimizers(self) :
        return torch.optim.Adam(self.parameters(), lr=self.lr)

    def save_huggingface_model(self, save_dir):
        """Will save the model, so you can reload it using `from_pretrained()`."""
        save_path = Path(save_dir)
        if not save_path.exists():
            save_path.mkdir(parents=True)
        state_dict = self.model.state_dict()
        torch.save(state_dict, os.path.join(save_dir, 'pytorch_model.bin'))
        with open(os.path.join(save_dir, 'config.json'), 'w', encoding='utf-8') as f:
            json.dump(self.config.to_dict(), f)



import requests

import pytest

from srtk.knowledge_graph import Freebase

FREEBASE_ENDPOINT = "http://localhost:3001/sparql"


def endpoint_reachable():
    try:
        response = requests.get(FREEBASE_ENDPOINT)
        if response.status_code == 200:
            return True
        else:
            return False
    except requests.exceptions.RequestException:
        return False

skip_if_unreachable = pytest.mark.skipif(
    not endpoint_reachable(), reason="Freebase endpoint not reachable"
)


@pytest.fixture
def freebase():
    return Freebase(FREEBASE_ENDPOINT, prepend_prefixes=True)

@skip_if_unreachable
def test_get_label(freebase: Freebase):
    # Skip the test if the fixture is skipped
    label = freebase.get_label("m.03_r3")
    assert label == "Jamaica"

@skip_if_unreachable
def test_search_one_hop_relations(freebase: Freebase):
    src = "m.03_r3"
    dst = "m.01428y"
    relations = freebase.search_one_hop_relations(src, dst)
    assert ["location.country.languages_spoken"] in relations, "(Jamaica --speak--> Jamaican English)"

@skip_if_unreachable
def test_search_two_hop_relations(freebase: Freebase):
    src = "m.0157m"
    dst = "m.0crpbj"
    relations = freebase.search_two_hop_relations(src, dst)
    assert ["people.person.education", "education.education.institution"] in relations, \
        "Bill Clinton --education-->  ?? --institution--> Hot Springs High School"

@skip_if_unreachable
def test_deduce_leaves(freebase: Freebase):
    src = "m.0157m"

    # Test one-hop leaves
    path = ("people.person.education",)
    leaves = freebase.deduce_leaves(src, path, limit=10)
    assert "m.0125cddf"  in leaves, "(Bill Clinton --education-->  )"
    
    # Test two-hop leaves
    path = ("people.person.education", "education.education.institution")
    leaves = freebase.deduce_leaves(src, path, limit=10)
    assert "m.0crpbj" in leaves, "(Bill Clinton --education-->  --insittution--> Hot Springs High School"

@skip_if_unreachable
def test_get_neighbor_relations(freebase):
    src = "m.0157m"
    # Test 1-hop relations
    relations = freebase.get_neighbor_relations(src, hop=1, limit=200)
    assert "government.us_president.vice_president" in relations and "government.politician.party" in relations,\
        "Billclinton has 1-hop relations vice_president and party"

    # Test 2-hop relations
    src = "m.0157m"
    relations = freebase.get_neighbor_relations(src, hop=2, limit=200)
    assert ('government.us_president.vice_president', 'people.person.gender') in relations,\
        "Bill Clinton --vice president--> --gender--> ..."

@skip_if_unreachable
def test_deduce_leaves_from_multiple_srcs(freebase: Freebase):
    srcs = ["m.0157m", "m.02mjmr"]  # Bill and Barack

    # Test searching one-hop common leaves
    path = ("government.us_president.vice_president",)
    leaves = freebase.deduce_leaves_from_multiple_srcs(srcs, path, limit=10)
    assert "m.0d05fv" in leaves, "(Bill | Barack --vice president--> Joe Biden )"



"""3. Search Path

This corresponds to search_to_get_path.py in the RUC's code.
It enumerates all paths from the question entities to answer entities.

python preprocess/search_path.py --ground-path data/preprocess/merged-ground.jsonl --output-path data/preprocess/paths.jsonl --remove-sample-without-path
"""
import sys
import os
import argparse

import srsly
from tqdm import tqdm

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', )))
from knowledge_graph import Wikidata, Freebase, KnowledgeGraphBase


def generate_paths(src_entities, dst_entities, kg: KnowledgeGraphBase, max_path=50):
    """Generate paths from question entities to answer entities.
    """
    paths = []
    for src in src_entities:
        for dst in dst_entities:
            if len(paths) >= max_path:
                break
            one_hop_paths = kg.search_one_hop_relations(src, dst)
            paths.extend(one_hop_paths)
            # If there are alreay one-hop paths between the two entities, 
            # we don't need to look further.
            if len(one_hop_paths) == 0:
                paths.extend(kg.search_two_hop_relations(src, dst))
    paths = [tuple(path) for path in paths]
    paths = list(set(paths))
    return paths[:max_path]


def has_type_relation(path):
    """A utility function to check whether the path contain certain relations."""
    for rel in path:
        if rel in ('type.object.type', 'type.type.instance'):
            return False
    return True


def main(args):
    # Each ground sample has the following fields:
    # - id: sample id
    # - question: question text
    # - question_entities: list of question entities
    # - answer_entities: list of answer entities
    ground_samples = srsly.read_jsonl(args.ground_path)
    total_samples = sum(1 for _ in srsly.read_jsonl(args.ground_path))
    if args.knowledge_graph == 'freebase':
        kg = Freebase(args.sparql_endpoint)
    else:
        kg = Wikidata(args.sparql_endpoint)
    processed_samples = []
    skipped = 0
    for sample in tqdm(ground_samples, total=total_samples, desc='Searching paths'):
        question_entities = sample['question_entities']
        answer_entities = sample['answer_entities']
        try:
            paths = generate_paths(question_entities, answer_entities, kg)
        except Exception as e:
            skipped += 1
            print(e)
            continue
        if args.remove_sample_without_path and not paths:
            skipped += 1
            continue
        # Special filter for Freebase
        if args.knowledge_graph == 'freebase':
            
            paths = list(filter(has_type_relation, paths))
        sample['paths'] = paths
        processed_samples.append(sample)
    print(f'Processed {len(processed_samples)} samples; skipped {skipped} samples without any paths between question entities and answer entities; total {total_samples} samples')
    srsly.write_jsonl(args.output_path, processed_samples)
    print(f'Retrieved paths saved to {args.output_path}')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--sparql-endpoint', default='http://localhost:1234/api/endpoint/sparql', help='knowledge graph SPARQL endpoint')
    parser.add_argument('--knowledge-graph', type=str, default='wikidata', choices=('wikidata', 'freebase'), help='knowledge graph name (default: wikidata)')
    parser.add_argument('--ground-path', type=str, required=True, help='grounded file where the question and answer entities are stored')
    parser.add_argument('--output-path', type=str, required=True, help='path file where several paths for each sample stored')
    parser.add_argument('--remove-sample-without-path', action='store_true', help='remove samples without paths')
    args = parser.parse_args()

    main(args)



from functools import lru_cache

import torch
from transformers import AutoTokenizer

from .encoder import LitSentenceEncoder


class Scorer:
    """Scorer for relation paths."""

    def __init__(self, pretrained_name_or_path, device=None):
        self.model = LitSentenceEncoder(pretrained_name_or_path)
        config = self.model.config
        self.tokenizer = AutoTokenizer.from_pretrained(config._name_or_path)
        if device:
            self.model = self.model.to(device)

    @lru_cache
    def score(self, question, prev_relations, next_relation):
        """Score a relation path.

        Args:
            question (str): question
            prev_relations (tuple[str]): tuple of relation **labels** that have been traversed.
            next_relation (str): next relation *label* to be traversed

        Returns:
            similarity: similarity between the query and the candidate
        """
        # Prepending 'query' and 'relation' corresponds to the way the model was trained (check collate_fn)
        query = f"query: {question} [SEP] {' # '.join(prev_relations)}"
        next_relation = 'relation: ' + next_relation
        text_pair = [query, next_relation]
        inputs = self.tokenizer(text_pair, return_tensors='pt', padding=True)
        inputs = {k: v.to(self.model.device) for k, v in inputs.items()}
        with torch.no_grad():
            outputs = self.model(**inputs, return_dict=True)
            query_embedding = outputs.last_hidden_state[0:1]
            sample_embedding = outputs.last_hidden_state[1:2]
            similarity = self.model.compute_sentence_similarity(
                query_embedding, sample_embedding)
        return similarity.item()

    @lru_cache
    def batch_score(self, question, prev_relations, next_relations):
        """Score next relations in batch.

        Args:
            question (str): question
            prev_relations (tuple[str]): tuple of relation **labels** that have been traversed.
            next_relations (tuple[str]): tuple of candidate next relation *labels* that are
                pertinent to the question and the previous relations.

        Returns:
            similarities (list[float]): list of similarities between the query and each candidate
        """
        query = f"query: {question} [SEP] {' # '.join(prev_relations)}"
        next_relations = ['relation: ' + next_relation for next_relation in next_relations]
        text_pair = [query] + next_relations
        inputs = self.tokenizer(text_pair, return_tensors='pt', padding=True)
        inputs = {k: v.to(self.model.device) for k, v in inputs.items()}
        with torch.no_grad():
            outputs = self.model(**inputs, return_dict=True)
            query_embedding = outputs.last_hidden_state[0:1]
            sample_embeddings = outputs.last_hidden_state[1:]
            similarities = self.model.compute_sentence_similarity(
                query_embedding, sample_embeddings)
            similarities = similarities.view(-1).tolist()
            if len(similarities) != len(next_relations):
                raise ValueError(f"Sanity check failed: {len(similarities)} != {len(next_relations)}")
        return similarities



"""1. Load the dataset
This scripts provides an example of how to prepare the dataset. It filters the grounded samples,
removing those without any answer or question entity.

The example input contains the following fields:
- id: the sample id
- sent: the question
- qc: the question entities
- ac: the answer entities

Example usage:

- For Mintaka dataset, those without any answer entity are removed.
$ python preprocess/load_dataset.py --dataset mintaka --ground-path data/preprocess/mintaka-ground-raw.jsonl --output-path data/preprocess/mintaka-ground.jsonl
Train + Validation + Test: Processed 12188 samples, skipped 7812 samples, total 20000 samples

- For MKQA dataset, those without any question entity are removed.
$ python preprocess/load_dataset.py --dataset mkqa --ground-path data/preprocess/mkqa-ground-raw.jsonl --output-path data/preprocess/mkqa-ground.jsonl
Processed 2112 samples, skipped 7888 samples, total 10000 samples
"""
import argparse
from pathlib import Path

import srsly
from tqdm import tqdm

def main(args):
    samples= srsly.read_jsonl(args.ground_path)
    total_lines = sum(1 for _ in srsly.read_jsonl(args.ground_path))
    skipped = 0
    processed_samples = []
    for sample in tqdm(samples, total=total_lines):
        if len(sample['qc']) == 0 or len(sample['ac']) == 0 or None in sample['ac']:
            skipped += 1
            continue
        processed_sample = {
            'id': args.dataset + '_' + str(sample['id']),
            'question': sample['sent'],
            'question_entities': sample['qc'],
            'answer_entities': sample['ac'],
        }
        processed_samples.append(processed_sample)

    output_path = Path(args.output_path)
    if not output_path.parent.exists():
        output_path.parent.mkdir(parents=True)
    srsly.write_jsonl(output_path, processed_samples)
    print(f'Processed {len(processed_samples)} samples, skipped {skipped} samples, total {total_lines} samples')
    print(f'Output saved to {output_path}')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--dataset', type=str, default='', help='dataset name, which will be prepend to sample ids')
    parser.add_argument('--ground-path', type=str, required=True)
    parser.add_argument('--output-path', type=str, required=True)
    args = parser.parse_args()
    main(args)



"""4. Score path

The score of a relation path is defined as the HIT rate of the prediction
with the ground truth entities. The *prediction* refers to the search results
from the question entities following the relation path.

Personal notes:
Why this is necessary? Isn't the relation path already the path from the question
entities to the ground truth entities?
In my understanding, this is similar to TF-IDF, the path is more precise if the results is
a smaller set of entities but have a higher intersection with the ground truth entities.

e.g.
python preprocess/score_path.py --paths-file data/preprocess/paths.jsonl --output-path data/preprocess/paths_scored.jsonl
"""
import os
import sys
import argparse

import srsly
from tqdm import tqdm

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', )))
from knowledge_graph import Wikidata, Freebase, KnowledgeGraphBase


def score_path(kg: KnowledgeGraphBase, src, path, answers, metric='jaccard'):
    """Calculate the HIT score of a given path.
    
    Args:
        kg (KnowledgeGraphBase): knowledge graph instance
        src (str): the source entity
        path (list): the path
        answers (list): the ground truth entities
        metric (str): how the paths are scored. 'jaccard' or 'recall'
            Default: 'jaccard', per the original implementation
    """
    if metric not in ('jaccard', 'recall'):
        raise ValueError(f'Unknown metric: {metric}')
    leaves = kg.deduce_leaves(src, path)
    leaves = set(leaves)
    answers = set(answers)
    hit = leaves.intersection(answers)
    if not leaves:
        # In the original implementation, they return 1 if the leaves is empty
        # I think it's more meaningful to set it to 0, as this path leads to no results
        return 0
    if metric == 'jaccard':
        score = len(hit) / len(leaves)
    else: # metric == 'recall':
        score = len(hit) / len(answers)
    return score


def main(args):
    if args.knowledge_graph == 'freebase':
        kg = Freebase(args.sparql_endpoint)
    else:
        kg = Wikidata(args.sparql_endpoint)
    samples = srsly.read_jsonl(args.paths_file)
    total_lines = sum(1 for _ in srsly.read_jsonl(args.paths_file))
    processed_samples = []  # adds path_scores to each sample
    # Each sample is a dict with the following fields:
    # - id: sample id
    # - question: question text
    # - question_entities: list of question entities
    # - answer_entities: list of answer entities
    # - paths: list of paths
    for sample in tqdm(samples, total=total_lines, desc='Scoring paths'):
        question_entities = sample['question_entities']
        answer_entities = sample['answer_entities']
        paths = sample['paths']
        path_scores = []
        for path in paths:
            # path score is the max score of all possible source entities following the path
            # Personal note: this is weird, why don't you start from the question entity where the
            # path was originally found?
            path = tuple(path)  # this makes it hashable
            score = max(score_path(kg, src, path, answer_entities, metric=args.metric) for src in question_entities)
            path_scores.append(score)
        sample['path_scores'] = path_scores
        processed_samples.append(sample)
    srsly.write_jsonl(args.output_path, processed_samples)
    print(f'Scored paths saved to {args.output_path}')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--sparql-endpoint', default='http://localhost:1234/api/endpoint/sparql', help='knowledge graph endpoint')
    parser.add_argument('-kg', '--knowledge-graph', default='wikidata', help='knowledge graph name')
    parser.add_argument('--paths-file', help='the file where the paths are stored')
    parser.add_argument('--output-path', help='the file where the scores are stored')
    parser.add_argument('--metric', default='jaccard', choices=('jaccard', 'recall'), help='the metric used to score the paths')
    args = parser.parse_args()

    main(args)



import pytest

from srtk.knowledge_graph import Wikidata

WIKIDATA_ENDPOINT = "https://query.wikidata.org/sparql"

@pytest.fixture
def wikidata():
    return Wikidata(WIKIDATA_ENDPOINT, prepend_prefixes=False, exclude_qualifiers=True)

def test_get_label(wikidata: Wikidata):
    label = wikidata.get_label("Q157808")
    assert label == "Technical University of Munich"

def test_search_one_hop_relations(wikidata: Wikidata):
    src = "Q157808"
    dst = "Q183"
    relations = wikidata.search_one_hop_relations(src, dst)
    assert ["P17"] in relations, "(TUM --country--> Germany)"

def test_search_two_hop_relations(wikidata: Wikidata):
    src = "Q157808"
    dst = "Q458"
    relations = wikidata.search_two_hop_relations(src, dst)
    assert ["P17", "P361"] in relations, "(TUM --country--> Germany --part of--> European Union)"

def test_deduce_leaves(wikidata: Wikidata):
    src = "Q157808"

    # Test one-hop leaves
    path = ("P17",)
    leaves = wikidata.deduce_leaves(src, path, limit=10)
    assert "Q183"  in leaves, "(TUM --country--> Germany )"
    
    # Test two-hop leaves
    path = ("P17", "P361")
    leaves = wikidata.deduce_leaves(src, path, limit=10)
    assert "Q458" in leaves and "Q27509" in leaves, "(TUM --country--> Germany --part of--> European Union | Central Europe)"

def test_get_neighbor_relations():
    src = "Q157808"
    wikidata = Wikidata(WIKIDATA_ENDPOINT, exclude_qualifiers=True)

    # Test 1-hop relations AND exclude_qualifiers=True
    relations = wikidata.get_neighbor_relations(src, hop=1, limit=200)
    assert "P17" in relations and "P112" in relations,  "TUM has 1-hop relations country and founded by"
    assert "P571" not in relations, "inception should be filtered out as it is a quantifier"

    # Test 2-hop relations
    relations = wikidata.get_neighbor_relations(src, hop=2)
    assert ("P17", "P361") in relations, "TUM --country--> Germany --part of--> European Union"

    # The wikidata endpoint is re-initialized to avoid caching
    wikidata = Wikidata(WIKIDATA_ENDPOINT, prepend_prefixes=False, exclude_qualifiers=False)
    # Test exclude_qualifiers=False
    relations = wikidata.get_neighbor_relations(src, hop=1, limit=200)
    assert "P571" in relations, "inception should be present when qualifiers are not excluded"

def test_deduce_leaves_from_multiple_srcs(wikidata: Wikidata):
    srcs = ["Q157808", "Q55044"]  # TUM and LMU

    # Test searching one-hop common leaves
    path = ("P17",)
    leaves = wikidata.deduce_leaves_from_multiple_srcs(srcs, path, limit=10)
    assert "Q183" in leaves, "(TUM | LMU --country--> Germany )"



import pytest
import torch

from srtk.scorer import LitSentenceEncoder


@pytest.fixture
def model():
    return LitSentenceEncoder('smallbenchnlp/roberta-small')


def test_pooler(model):
    if hasattr(model, 'average_pool'):
        pooler = model.average_pool
    elif hasattr(model, 'cls_pool'):
        pooler = model.cls_pool
    else:
        pytest.skip('No pooler found')
    # Test pooler without batch size
    hidden_states = torch.randn(10, 256)
    attention_mask = torch.ones(10)
    pooled = pooler(hidden_states, attention_mask)
    assert pooled.shape == (256,)
    # Test pooler with batch size
    hidden_states = torch.randn(2, 10, 256)
    attention_mask = torch.ones(2, 10)
    pooled = pooler(hidden_states, attention_mask)
    assert pooled.shape == (2, 256)

def test_compute_sentence_similarity(model):
    # Test batched similarity
    query = torch.randn(2, 1, 10, 256)
    samples = torch.randn(2, 3, 10, 256)
    similarity = model.compute_sentence_similarity(query, samples)
    assert similarity.shape == (2, 3)
    # Test single pair similarity
    query = torch.randn(1, 10, 256)
    sample = torch.randn(1, 10, 256)
    similarity = model.compute_sentence_similarity(query, sample)
    assert similarity.shape == (1,)

def test_training_step(model):
    # Test training step
    batch = {
        'input_ids': torch.randint(0, 100, (2, 4, 10)),
        'attention_mask': torch.ones(2, 4, 10),
    }
    loss = model.training_step(batch, 0)
    assert loss.shape == ()



"""The script to train the scorer model.

e.g.
python train.py --data-file data/train.jsonl --model-name-or-path intfloat/e5-small --save-model-path artifacts/scorer
"""
import argparse
import datetime
from collections import defaultdict
from dataclasses import dataclass

import lightning.pytorch as pl
import torch
from datasets import load_dataset
from lightning.pytorch.loggers import WandbLogger
from torch.utils.data import DataLoader
from transformers import AutoTokenizer, PreTrainedTokenizerBase

from .scorer import LitSentenceEncoder


def concate_all(example):
    """Concatenate all columns into one column for input.
    The resulted 'input_text' column is a list of strings.
    """
    query = 'query: ' + example['query']
    rels = [example['positive']] + example['negatives']
    rels = ['relation: ' + rel for rel in rels]
    example['input_text'] = [query] + rels
    return example


@dataclass
class Collator:
    """Collate a list of examples into a batch."""
    tokenizer: PreTrainedTokenizerBase

    def __call__(self, features):
        batched = defaultdict(list)
        for item in features:
            for key, value in item.items():
                value = torch.tensor(value)
                if key == 'attention_mask':
                    value = value.bool()
                batched[key].append(value)
        for key, value in batched.items():
            batched[key] = torch.stack(value, dim=0)
        return batched


def prepare_dataloaders(train_data, validation_data, tokenizer, batch_size):
    """Prepare dataloaders for training and validation."""
    def tokenize(example):
        tokenized = tokenizer(example['input_text'], padding='max_length', truncation=True, return_tensors='pt', max_length=32)
        return tokenized

    train_split = 'train[:95%]' if validation_data is None else 'train'
    validation_split = 'train[95%:]' if validation_data is None else 'train'
    if validation_data is None:
        validation_data = train_data

    train_dataset = load_dataset('json', data_files=train_data, split=train_split)
    train_dataset = train_dataset.map(concate_all, remove_columns=train_dataset.column_names)
    train_dataset = train_dataset.map(tokenize, remove_columns=train_dataset.column_names)
    validation_dataset = load_dataset('json', data_files=validation_data, split=validation_split)
    validation_dataset = validation_dataset.map(concate_all, remove_columns=validation_dataset.column_names)
    validation_dataset = validation_dataset.map(tokenize, remove_columns=validation_dataset.column_names)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, collate_fn=Collator(tokenizer), num_workers=8)
    validation_loader = DataLoader(validation_dataset, batch_size=batch_size, shuffle=False, collate_fn=Collator(tokenizer), num_workers=8)
    return train_loader, validation_loader


def main(args):
    torch.set_float32_matmul_precision('medium')
    model = LitSentenceEncoder(args.model_name_or_path, lr=args.learning_rate)
    tokenizer = AutoTokenizer.from_pretrained(args.model_name_or_path)
    train_loader, validation_loader = prepare_dataloaders(args.train_dataset, args.validation_dataset, tokenizer, args.batch_size)
    day_hour = datetime.datetime.now().strftime('%m%d%H%M')
    wandb_logger = WandbLogger(project='retrieval', name=day_hour , group='contrastive', save_dir='artifacts')
    trainer = pl.Trainer(accelerator=args.accelerator, default_root_dir=args.output_dir,
                         fast_dev_run=args.fast_dev_run, max_epochs=args.max_epochs, logger=wandb_logger)
    trainer.fit(model, train_dataloaders=train_loader, val_dataloaders=validation_loader)
    model.save_huggingface_model(args.output_dir)
    tokenizer.save_pretrained(args.output_dir)


def add_arguments(parser):
    """Add train arguments to a parser in place."""
    parser.add_argument('-t', '--train-dataset', required=True,
                        help='path to the training dataset. It should be a JSONL file with fields: query, positive, negatives')
    parser.add_argument('-v', '--validation-dataset',
                        help='path to the validation dataset. If not provided, 5 percent of the training data will be used as validation data.\
                        (default: None)')
    parser.add_argument('-o', '--output-dir', default='artifacts/scorer',
                        help='output model path. the model will be saved in the format of huggingface models,\
                        which can be uploaded to the huggingface hub and shared with the community.\
                        (default: artifacts/scorer)')
    parser.add_argument('--model-name-or-path', default='intfloat/e5-small',
                        help='pretrained model name or path. It is fully compatible with HuggingFace models.\
                        You can specify either a local path where a model is saved, or an encoder model identifier\
                        from huggingface hub. (default: intfloat/e5-small)')
    parser.add_argument('-lr', '--learning-rate', default=5e-5, type=float, help='learning rate (default: 5e-5)')
    parser.add_argument('--batch-size', default=16, type=int, help='batch size (default: 16)')
    parser.add_argument('--max-epochs', default=10, type=int, help='max epochs (default: 10)')
    parser.add_argument('--accelerator', default='gpu', help='accelerator, can be cpu, gpu, or tpu (default: gpu)')
    parser.add_argument('--fast-dev-run', action='store_true',
                        help='fast dev run for debugging, only use 1 batch for training and validation')

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()

    main(args)



"""This script retrieves subgraphs from a knowledge graph according to a natural
language query (usually a question). This command can also be used to evaluate
a trained retriever when the answer entities are known.

The expected fields of one sample are:
- question: question text
- question_entities: list of grounded question entities (ids)

For evaluation, the following field is also required:
- answer_entities: list of grounded answer entities (ids)
"""
import argparse
import heapq
import os
import pathlib
from collections import namedtuple
from typing import List, Any, Dict

import srsly
import torch
from tqdm import tqdm

from .knowledge_graph import Freebase, Wikidata, KnowledgeGraphBase
from .scorer import Scorer


END_REL = 'END OF HOP'


# Path collects the information at each traversal step
# - prev_relations stores the relations that have been traversed
# - score stores the score of the relation path
Path = namedtuple('Path', ['prev_relations', 'score'], defaults=[(), 0])


class KnowledgeGraphTraverser:
    '''KnowledgeGraphTraverser is a helper class that traverses a knowledge graph'''
    def __init__(self, kg: KnowledgeGraphBase):
        self.kg = kg

    def retrive_subgraph(self, entity, path):
        '''Retrive subgraph entities and triplets by traversing from an entity following
        a relation path hop by hop.

        Args:
            entity: the source node
            path: a list of relations

        Returns:
            entities: a list of entities
            triplets: a list of triplets
        '''
        entities, triplets = set(), set()
        tracked_entities = set((entity,))
        for relation in path:
            next_hops = set()
            if relation == END_REL:
                continue
            for e in tracked_entities:
                leaves = set(self.kg.deduce_leaves(e, (relation,)))
                next_hops |= leaves
                triplets |= {(e, relation, leaf) for leaf in leaves}
            entities |= next_hops
            tracked_entities = next_hops
        return list(entities), list(triplets)

    def deduce_leaves(self, entity, path):
        """Deduce leaves from an entity following a path hop by hop.

        Args:
            entity: the source node
            path: a list of relations

        Returns:
            leaves: a list of leaves that are n-hop away from the source node,
                where n is the length of the path
        """
        leaves = set((entity,))
        for relation in path:
            if relation == END_REL:
                continue
            leaves = set().union(*(self.kg.deduce_leaves(leaf, (relation,)) for leaf in leaves))
        return leaves

    def deduce_leaf_relations(self, entity, path):
        """Deduce leaf relations from an entity following a path hop by hop.

        Args:
            entity: the source node
            path: a list of relations

        Returns:
            relations: a list of relations that are n-hop away from the source node,
                where n is the length of the path
        """
        leaves = self.deduce_leaves(entity, path)
        relations = set().union(*(self.kg.get_neighbor_relations(leaf) for leaf in leaves))
        # Special filter relation for freebase
        if self.kg.name == 'freebase':
            relations = [r for r in relations if r.split('.')[0] not in ['kg', 'common']]
        return tuple(relations)


class Retriever:
    '''Retriever retrieves subgraphs from a knowledge graph with a question and its
    linked entities. The retrieval process takes the semantic information of the question
    and the expanding path into consideration.
    '''
    def __init__(self, kg: KnowledgeGraphBase, scorer: Scorer, beam_width: int, max_depth: int):
        self.kgh = KnowledgeGraphTraverser(kg)
        self.scorer = scorer
        self.beam_width = beam_width
        self.max_depth = max_depth
        self.num_entity_threshold = 1000

    def retrieve_subgraph_triplets(self, sample: Dict[str, Any]):
        """Retrieve triplets as subgraphs from paths.

        Args:
            sample: a sample from the dataset, which contains at least the following fields:
                question: a string
                question_entities: a list of entities

        Returns:
            triplets: a list of triplets
        """
        question = sample['question']
        triplets = []
        for question_entity in sample['question_entities']:
            path_score_list = self.beam_search_path(question, question_entity)
            n_nodes = 0
            for relations, _ in path_score_list:
                partial_nodes, partial_triples = self.kgh.retrive_subgraph(question_entity, relations)
                if len(partial_nodes) > self.num_entity_threshold:
                    continue
                n_nodes += len(partial_nodes)
                triplets.extend(partial_triples)

                if n_nodes > self.num_entity_threshold:
                    break
        triplets = list(set(triplets))
        return triplets

    def beam_search_path(self, question: str, question_entity: str):
        '''This function reimplement RUC's paper's solution. In the search process, only the history
        paths are recorded; each new relation is looked up via looking up the end relations from the
        question entities following a history path.

        Args:
            question (str)
            question_entity (str): a grounded question entity

        Returns:
            path_score_list (list[Path]): a list of (path, score) tuples
        '''
        candidate_paths = [Path()]  # path and its score
        result_paths = []
        depth = 0

        while candidate_paths and len(result_paths) < self.beam_width and depth < self.max_depth:
            next_relations_batched = []
            for path in candidate_paths:
                prev_relations = path.prev_relations
                next_relations = self.kgh.deduce_leaf_relations(question_entity, prev_relations)
                next_relations = next_relations + (END_REL,)
                next_relations_batched.append(next_relations)

            tracked_paths = self.expand_and_score_paths(question, candidate_paths, next_relations_batched)
            tracked_paths = heapq.nlargest(self.beam_width, tracked_paths, key=lambda x: x.score)
            depth += 1
            # Update candidate_paths
            candidate_paths = []
            for path in tracked_paths:
                if  path.prev_relations and path.prev_relations[-1] == END_REL:
                    result_paths.append(path)
                else:
                    candidate_paths.append(path)
        # Merge not-yet-ended paths into the result paths 
        candidate_paths = [Path(prev_relations + (END_REL,), score) for prev_relations, score in candidate_paths]
        result_paths = heapq.nlargest(self.beam_width, result_paths + candidate_paths, key=lambda x: x.score)
        return result_paths

    def expand_and_score_paths(self, question: str, paths: List[Path], relations_batched: List[List[str]]) -> List[Path]:
        '''Expand the paths by one hop and score them by comparing the embedding similarity between
        the query (question + prev_relations) and the next relation.

        Args:
            question (str)
            paths (list[Path]): a list of current paths
            relations_batched (list[list[str]]): a list of next relations for each path

        Returns:
            scored_paths (list[Path]): a list of newly expanded and scored paths
        '''
        scored_paths = []
        score_matrix = []
        for path, next_relations in zip(paths, relations_batched):
            scores = self.scorer.batch_score(question, tuple(path.prev_relations), tuple(next_relations))
            score_matrix.append(scores)

        for i, (path, next_relations) in enumerate(zip(paths, relations_batched)):
            for j, relation in enumerate(next_relations):
                new_prev_relations = path.prev_relations + (relation,)
                score = float(score_matrix[i][j]) + path.score
                scored_paths.append(Path(new_prev_relations, score))
        return scored_paths


def calculate_hit_and_miss(retrieved_path):
    """Calculate the recall of answer entities in retrieved triplets,
    if answer_entities exists in each sample.
    """
    retrieval = srsly.read_jsonl(retrieved_path)
    hit = 0
    miss = 0
    for sample in retrieval:
        if 'answer_entities' not in sample:
            continue
        answers = sample['answer_entities']
        entities = set().union(*((triplet[0], triplet[-1]) for triplet in sample['triplets']))
        if any([entity in answers for entity in entities]):
            hit += 1
        else:
            miss += 1
    return hit, miss


def print_and_save_recall(retrieved_path):
    """Calculate and print the recall of answer entities in retrieved triplets,
    If any answer from the answer entities is in the retrieved entities, the sample
    counts as a hit.
    """
    hit, miss = calculate_hit_and_miss(retrieved_path)
    print(f"Answer recall: {hit / (hit + miss)} ({hit} / {hit + miss})")
    info = {}
    if hit + miss != 0:
        info = {
            'hit': hit,
            'miss': miss,
            'recall': hit / (hit + miss)
        }
    # path/to/subgraph.jsonl -> path/to/subgraph.metric
    recall_path = os.path.splitext(retrieved_path)[0] + '.metric'
    srsly.write_json(recall_path, info)


def main(args):
    pathlib.Path(os.path.dirname(args.output)).mkdir(parents=True, exist_ok=True)
    if args.knowledge_graph == 'freebase':
        kg = Freebase(args.sparql_endpoint)
    else:
        kg = Wikidata(args.sparql_endpoint, exclude_qualifiers=not args.include_qualifiers)
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    scorer = Scorer(args.scorer_model_path, device)
    retriever = Retriever(kg, scorer, args.beam_width, args.max_depth)
    samples = list(srsly.read_jsonl(args.input))
    total = sum(1 for _ in srsly.read_jsonl(args.input))
    for sample in tqdm(samples, desc='Retrieving subgraphs', total=total):
        triplets = retriever.retrieve_subgraph_triplets(sample)
        sample['triplets'] = triplets
    srsly.write_jsonl(args.output, samples)
    print(f'Retrieved subgraphs saved to to {args.output}')
    if args.evaluate:
        print_and_save_recall(args.output)


def add_arguments(parser):
    """Add retrieve arguments to the parser in place."""
    parser.description = '''Retrieve subgraphs with a trained model on a dataset that entities are linked.
    This command can also be used to evaluate a trained retriever when the answer entities are known.

    Provide a JSON file as input, where each JSON object must contain at least the 'question' and 'question_entities' fields.
    When ``--evaluate`` is set, the input JSON file must also contain the 'answer_entities' field.

    The output JSONL file will include an added 'triplets' field, based on the input JSONL file. This field consists of a list of triplets,
    with each triplet representing a (head, relation, tail) tuple.
    When ``--evaluate`` is set, a metric file will also be saved to the same directory as the output JSONL file.
    '''
    parser.add_argument('-i', '--input', type=str, required=True, help='path to input jsonl file. it should contain at least \
                        ``question`` and ``question_entities`` fields.')
    parser.add_argument('-o', '--output', type=str, required=True, help='output file path for storing retrieved triplets.')
    parser.add_argument('-e', '--sparql-endpoint', type=str, help='SPARQL endpoint for Wikidata or Freebase services.')
    parser.add_argument('-kg', '--knowledge-graph', type=str, required=True, choices=('freebase', 'wikidata'),
                        help='choose the knowledge graph: currently supports ``freebase`` and ``wikidata``.')
    parser.add_argument('--scorer-model-path', type=str, required=True, help='Path to the scorer model, containing \
                        both the saved model and its tokenizer in the Huggingface models format.\
                        Such a model is saved automatically when using the ``srtk train`` command.\
                        Alternatively, provide a pre-trained model name from the Hugging Face model hub.\
                        In practice it supports any Huggingface transformers encoder model, though models that do not use [CLS] \
                        tokens may require modifications on similarity function.')
    parser.add_argument('--beam-width', type=int, default=10, help='beam width for beam search (default: 10).')
    parser.add_argument('--max-depth', type=int, default=2, help='maximum depth for beam search (default: 2).')
    parser.add_argument('--evaluate', action='store_true', help='Evaluate the retriever model. When the answer \
                        entities are known, the recall can be evluated as the number of samples that any of the \
                        answer entities are retrieved in the subgraph by the number of all samples. This equires \
                        `answer_entities` field in the input jsonl.')
    parser.add_argument('--include-qualifiers', action='store_true', help='Include qualifiers from the retrieved triplets. \
                        Qualifiers are informations represented in non-entity form, like date, count etc.\
                        This is only relevant for Wikidata.')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()
    if not args.sparql_endpoint:
        if args.knowledge_graph == 'freebase':
            args.sparql_endpoint = 'http://localhost:3001/sparql'
        else:
            args.sparql_endpoint = 'http://localhost:1234/api/endpoint/sparql'
        print(f'Using default sparql endpoint for {args.knowledge_graph}: {args.sparql_endpoint}')
    main(args)



'''Provide protocal for different kinds of knowledge graphs.'''

from abc import abstractmethod
from typing import List


class KnowledgeGraphBase:
    """Knowledge graph base class."""

    @abstractmethod
    def search_one_hop_relations(self, src: str, dst: str) -> List[List[str]]:
        """Search one hop relations between src and dst.

        Args:
            src (str): source entity
            dst (str): destination entity

        Returns:
            list[list[str]]: list of paths, each path is a list of PIDs
        """
        raise NotImplementedError

    @abstractmethod
    def search_two_hop_relations(self, src: str, dst: str) -> List[List[str]]:
        """Search two hop relations between src and dst.

        Args:
            src (str): source entity
            dst (str): destination entity

        Returns:
            list[list[str]]: list of paths, each path is a list of PIDs
        """
        raise NotImplementedError

    @abstractmethod
    def deduce_leaves(self, src: str, path: List[str], limit: int) -> List[str]:
        """Deduce leave entities from source entity following the path.

        Args:
            src_entity (str): source entity
            path (tuple[str]): path from source entity to destination entity
            limit (int, optional): limit of the number of leaves.

        Returns:
            list[str]: list of leaves. Each leaf is a QID.
        """
        raise NotImplementedError

    @abstractmethod
    def get_neighbor_relations(self, src: str, hop: int, limit: int) -> List[str]:
        """Get relations between src and dst.

        Args:
            src (str): source entity
            hop (int, optional): hop of the relations. Defaults to 1.
            limit (int, optional): limit of the number of relations.

        Returns:
            list[str] | list[tuple[str]]: list of relations (one-hop)
                or list of tuples of relations (multi-hop)
        """
        raise NotImplementedError

    @abstractmethod
    def get_label(self, identifier: str) -> str:
        """Get label of an entity or a relation.

        Args:
            identifier (str): entity or relation identifier

        Returns:
            str: label of the entity or the relation
        """
        raise NotImplementedError

    def get_relation_label(self, relation: str) -> str:
        """Get the label of a relation. Defaults to get_label.

        Args:
            relation (str): relation identifier

        Returns:
            str: label of the relation
        """
        return self.get_label(relation)

    def get_entity_label(self, entity: str) -> str:
        """Get the label of an entity. Defaults to get_label.

        Args:
            entity (str): entity identifier

        Returns:
            str: label of the entity
        """
        return self.get_label(entity)



"""2. This script merges the grounded data into one training data.

e.g.
python preprocess/merge_ground.py  --output-path data/preprocess/merged-ground.jsonl --ground-files\
    data/preprocess/mintaka-ground.jsonl data/preprocess/mkqa-ground.jsonl
"""
import argparse
import srsly

def main(args):
    merged_samples = []
    for ground_file in args.ground_files:
        samples = srsly.read_jsonl(ground_file)
        merged_samples.extend(samples)
    srsly.write_jsonl(args.output_path, merged_samples)

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--output-path', type=str, required=True)
    parser.add_argument('--ground-files', nargs='+', required=True)
    args = parser.parse_args()
    main(args)



"""This script creates the training data from the grounded questions.

Inputs should be a jsonl file, with each line representing a grounded question.
The format of each line should be like this example:
```json
{
  "id": "sample-id",
  "question": "Which universities did Barack Obama graduate from?",
  "question_entities": [
    "Q76"
  ],
  "answer_entities": [
    "Q49122",
    "Q1346110",
    "Q4569677"
  ]
}
```
"""
import os
import argparse
from argparse import Namespace
from pathlib import Path

from .preprocessing.search_path import main as search_path
from .preprocessing.score_path import main as score_path
from .preprocessing.negative_sampling import main as negative_sampling


def main(args):
    output_path = args.output
    # Create parent dir for output if not exists.
    Path(os.path.dirname(output_path)).mkdir(parents=True, exist_ok=True)
    if args.search_path:
        intermediate_dir = args.intermediate_dir
        if intermediate_dir is None:
            intermediate_dir = os.path.dirname(output_path)
        Path(intermediate_dir).mkdir(parents=True, exist_ok=True)
        paths_file = os.path.join(intermediate_dir, 'paths.jsonl')
        scores_file = os.path.join(intermediate_dir, 'scores.jsonl')
        search_path_args = Namespace(sparql_endpoint=args.sparql_endpoint,
                                    knowledge_graph=args.knowledge_graph,
                                    ground_path=args.input,
                                    output_path=paths_file,
                                    remove_sample_without_path=True)
        score_path_args = Namespace(sparql_endpoint=args.sparql_endpoint,
                                    knowledge_graph=args.knowledge_graph,
                                    paths_file=paths_file,
                                    output_path=scores_file,
                                    metric=args.metric,)
        negative_sampling_args = Namespace(sparql_endpoint=args.sparql_endpoint,
                                           knowledge_graph=args.knowledge_graph,
                                           scored_path_file=scores_file,
                                           num_negative=args.num_negative,
                                           positive_threshold=args.positive_threshold,
                                           output_path=output_path,)
        search_path(search_path_args)
        score_path(score_path_args)
    else:
        negative_sampling_args = Namespace(sparql_endpoint=args.sparql_endpoint,
                                           knowledge_graph=args.knowledge_graph,
                                           scored_path_file=args.input,
                                           num_negative=args.num_negative,
                                           positive_threshold=args.positive_threshold,
                                           output_path=output_path,)
    negative_sampling(negative_sampling_args)


def add_arguments(parser):
    """Add preprocess arguments to a parser in place."""
    parser.description = 'Create the training data from the grounded questions.'
    parser.add_argument('-i', '--input', type=str, required=True,
                        help='The grounded questions file with question, question & answer entities')
    parser.add_argument('-o', '--output', type=str, required=True,
                        help='The output path where the final training data will be saved.')
    parser.add_argument('--intermediate-dir', type=str, help="The directory to save intermediate files. If not specified, the intermediate \
                        files will be saved in the same directory as the output file, with the name paths.jsonl and scores.jsonl")
    parser.add_argument('-e', '--sparql-endpoint', type=str, required=True,
                        help="SPARQL endpoint URL for either Wikidata or Freebase\
                        (e.g., 'http://localhost:1234/api/endpoint/sparql' for default local qEndpoint)")
    parser.add_argument('-kg', '--knowledge-graph', type=str, required=True, choices=('wikidata', 'freebase'),
                        help='knowledge graph name, either wikidata or freebase')
    parser.add_argument('--search-path', action='store_true',
                        help='Whether to search paths between question and answer entities. If not specified, paths and scores fields\
                        must present in the input file. You **have to** specify this for weakly supervised learning. (default: False)')
    parser.add_argument('--metric', choices=('jaccard', 'recall'), default='jaccard',
                        help='The metric used to score the paths. recall will usually result in a lager size of training dataset.\
                        (default: jaccard))')
    parser.add_argument('--num-negative', type=int, default=15,
                        help='The number of negative relations to sample for each positive relation. (default: 15)')
    parser.add_argument('--positive-threshold', type=float, default=0.5,
                        help='The threshold to determine whether a path is positive or negative. If you want to use \
                        a larger training dataset, you can set this value to a smaller value. (default: 0.5)')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()
    main(args)



import pytest
from srtk.scorer import Scorer


@pytest.fixture
def scorer():
    """Whether the scorer can be initialized"""
    return Scorer('smallbenchnlp/roberta-small')

def test_score(scorer):
    """Whether the scorer can run and return a digit, without crashing."""
    query = 'query: how old is the natural satellite of the earch?'
    prev_relations = ('satellite of',)
    next_relation = 'relation: age'
    scorer.score(query, prev_relations, next_relation)

def test_batch_score(scorer):
    """Whether the batch scorer is capable of batch scoring."""
    query = 'query: how old is the natural satellite of the earch?'
    prev_relations = ('satellite of',)
    next_relations = ('relation: age', 'relation: age')
    scores = scorer.batch_score(query, prev_relations, next_relations)
    assert isinstance(scores, list)
    assert len(scores) == len(next_relations)
    assert scores[0] == scores[1], 'The scores should be the same for the same samples.'
